<?php
namespace app\core\helpers;

use Yii;
use Exception;

class UpdateHelper
{
    public static function checkUpdates($verbose = false)
    {
        $rootDir = Yii::getAlias('@project');
        $version = shell_exec("cd $rootDir; git log --pretty=format:'%h' -n 1");

        if ($verbose) {
            echo 'Current version: ' . $version . PHP_EOL;
        }
        
        $status = shell_exec("cd $rootDir; git status");
        if ($verbose) {
            echo $status . PHP_EOL;
        }

        if (strpos($status, 'Changes not staged for commit') !== false || strpos($status, 'Changes to be committed') !== false) {
            if ($verbose) {
                echo "Untracked files present. Cannot continue." . PHP_EOL;
            }
            return false;
        }

        if (strpos($status, 'Your branch is up to date') === false && strpos($status, 'Your branch is behind') === false) {
            if ($verbose) {
                echo 'Unknown repository state. Stopping...' . PHP_EOL;
            }
            return false;
        }


        $serverResponse = file_get_contents(Yii::$app->params['updateUrl']);
        if (!$serverResponse) {
            return false;
        }
        $serverResponse = json_decode($serverResponse);
        if (!$serverResponse) {
            return false;
        }

        if ($serverResponse->version && $serverResponse->version != $version) {
            return $serverResponse->version;
        }

        return false;
    }

    public static function performUpdate($verbose = false)
    {
        $haveUpdates = static::checkUpdates($verbose);
        $rootDir = Yii::getAlias('@project');
        $version = shell_exec("cd $rootDir; git log --pretty=format:'%h' -n 1");

        if (!$haveUpdates) {
            return false;
        }

        Yii::$app->settings->set('app.core', 'updating-from', $version);

        $rootDir = Yii::getAlias('@project');

        $sshCommand = "ssh -o StrictHostKeyChecking=no -i $rootDir/plugins/app/update/keys/id_rsa_gitpf";
        $php = PHP_BINARY;
        if (strpos($php, 'php-fpm') !== false) {
            $php = isset(Yii::$app->params['phpBinary'])? Yii::$app->params['phpBinary'] : 'php';
        }

        $composer = trim(shell_exec('which composer'));

        $output = [];
        $code = null;
        $command = "bash -c 'cd $rootDir && GIT_SSH_COMMAND=\"$sshCommand\" && git pull && $php $composer install && $php yii platform/plugin/install-all'";
        if ($verbose) {
            echo $command;
        }

        exec($command, $output, $code);

        if ($code !== 0) {
            Yii::$app->settings->set('app.core', 'update-error', $output);
            static::reportUpdate($version, $haveUpdates, 0, $output);
            throw new Exception("Update failed: " . implode(PHP_EOL, $output));
        }

        if ($verbose) {
            echo "Update done" . PHP_EOL;
        }

        Yii::$app->settings->set('app.core', 'updated-to', shell_exec("cd $rootDir; git log --pretty=format:'%h' -n 1"));
        Yii::$app->cache->delete('app-updates');

        static::reportUpdate($version, $haveUpdates, 1);
        return true;
    }

    public static function reportUpdate($from, $to, $status, $error = '')
    {
        $url = Yii::$app->params['updateReportUrl'];
        $url = $url . '?' . http_build_query([
            'status' => $status,
            'message' => 'Update from ' . $from . ' to ' . $to . $error,
            'host' => Yii::getAlias('@frontendUrl'),
        ]);

        $answer = file_get_contents($url);

        if ($answer == 'OK') {
            return true;
        }
    }

    public static function getHistory($from, $to)
    {
        $from = escapeshellarg($from);
        $to = escapeshellarg($to);
        $history = shell_exec("cd " . Yii::getAlias('@project') . "; git log --pretty=oneline --abbrev-commit " . $from . "..." . $to);
        return $history;
    }

    public static function getUpdateChangeLog($from, $to)
    {
        if (!$from) {
            $rootDir = Yii::getAlias('@project');
            $from = shell_exec("cd $rootDir; git log --pretty=format:'%h' -n 1");
        }
        $serverResponse = file_get_contents(Yii::$app->params['updateChangeLogUrl'] . '?' . http_build_query(['from' => $from, 'to' => $to]));
        if (!$serverResponse) {
            return 'Не удалось получить историю изменений';
        }
        return $serverResponse;
    }

    public static function version()
    {
        $cache = Yii::$app->cache;
        return $cache->getOrSet('app-version', function () {
            $rootDir = Yii::getAlias('@project');
            return shell_exec("cd $rootDir; git log --pretty=format:'%h' -n 1");
        }, 300);
    }
}
