<?php

namespace app\gitlab\components;

use Yii;
use Exception;
use Throwable;
use yii\base\Component;
use app\core\models\Issue;
use app\core\models\IssueComment;

use Gitlab\Client;
use Gitlab\Model\User as GitlabUser;
use Gitlab\Model\Project as GitlabProject;
use Gitlab\Model\Issue as GitlabIssue;
use app\planfix\classes\PFApi;
use app\core\models\Project as AppProject;

use app\core\models\Gitlab;
use app\core\models\AppUser as User;
use League\HTMLToMarkdown\HtmlConverter;
use app\planfix\helpers\ApiHelper;
use app\core\models\CoreStatus;
use app\core\models\RequestLog;
use Gitlab\HttpClient\Builder;
use app\gitlab\utils\RequestLogger;
use yii\db\ActiveRecord;

/**
 * Class GitlabApi - загрузка данных из Гитлаба
 *
 * @package app\gitlab\components
 */
class GitlabApi extends Component
{
    /**
     * @var Client Client
     */
    public $client = null;

    /**
     * @var string $token Gitlab private token
     */
    public $token = null;

    /**
     * @var string $url
     */
    public $url = null;

    /**
     * @var bool $noHooks
     */
    public $noHooks = true;
    /**
     * @var bool
     */
    public $isAdmin = false;


    public function init()
    {
        parent::init();
        try {
            $gitlab = Gitlab::find()->one();
            $this->token = $gitlab? $gitlab->token : '';
            $this->url = $gitlab? $gitlab->url : '';

            if ($this->token && $this->url) {
                if ($this->url[mb_strlen($this->url) - 1] == '/') {
                    $this->url = substr($this->url, 0, -1);
                }

                $builder = new Builder();
                $builder->addPlugin(new RequestLogger());

                $this->client = new Client($builder);
                $this->client->setUrl($this->url);
                $this->client->authenticate($this->token, Client::AUTH_URL_TOKEN);

                $this->isAdmin = Yii::$app->cache->getOrSet('gl-isAdmin', function () {
                    $currentUser = $this->raw('/user');
                    return (property_exists($currentUser, 'is_admin') && $currentUser->is_admin)? true : false;
                }, 6000);
            }
        } catch (Exception $e) {
        }
    }

    /**
     *  Запись запроса в БД
     */
    public function logRequest()
    {
        $log = RequestLog::find()->where([
            'date' => date('Y-m-d'),
            'service' => 'gitlab'
        ])->one();

        if (!$log) {
            $log = new RequestLog();
            $log->service = 'gitlab';
            $log->count = 0;
            $log->date = date('Y-m-d');
        }

        $log->count += 1;
        $log->save();
    }

    /**
     * Импорт проекта из GitLab
     * @param $gitlabId
     * @return AppProject
     * @throws Exception
     */
    public function importProject($gitlabId)
    {
        $gitlabProject = new GitlabProject($gitlabId, $this->client);
        $gitlabProject = $gitlabProject->show();

        $project = AppProject::find()->where([
            'gitlab_id' => $gitlabId,
        ])->one();

        if (!$project) {
            $project = new AppProject();
            $project->gitlab_id = $gitlabId;
            $project->title = $gitlabProject->name;
            $project->description = $gitlabProject->description;
            
            if ($gitlabProject->owner) {
                $project->owner_id = $this->getUser($gitlabProject->owner->id)->id;
            }

            $project->save();
        }

        PFApi::syncProjectFromModel($project);
        PFApi::syncPlanfixProject($project->planfix_id);

        return $project;
    }

    /**
     * Очистка и транслитерация названия проекта
     * @param $string
     * @return string|string[]
     * @throws Exception
     */
    private function cleanProjectName($string)
    {
        $string = str_replace(' ', '-', $string); // Replaces all spaces with hyphens.

        $cyr = [
            'а','б','в','г','д','е','ё','ж','з','и','й','к','л','м','н','о','п',
            'р','с','т','у','ф','х','ц','ч','ш','щ','ъ','ы','ь','э','ю','я',
            'А','Б','В','Г','Д','Е','Ё','Ж','З','И','Й','К','Л','М','Н','О','П',
            'Р','С','Т','У','Ф','Х','Ц','Ч','Ш','Щ','Ъ','Ы','Ь','Э','Ю','Я'
        ];
        $lat = [
            'a','b','v','g','d','e','io','zh','z','i','y','k','l','m','n','o','p',
            'r','s','t','u','f','h','ts','ch','sh','sht','a','i','y','e','yu','ya',
            'A','B','V','G','D','E','Io','Zh','Z','I','Y','K','L','M','N','O','P',
            'R','S','T','U','F','H','Ts','Ch','Sh','Sht','A','I','Y','e','Yu','Ya'
        ];
        $string = str_replace($cyr, $lat, $string);

        $string = preg_replace('/[^A-Za-z0-9\-]/', '', $string); // Removes special chars.

        if (!$string) {
            throw new Exception("Неверное имя проекта");
        }

        return $string;
    }

    /**
     * Проврка проекта на удаление
     * @param AppProject $project
     * @return bool
     * @throws Throwable
     */
    public function checkProject(AppProject $project)
    {
        $gitlabProject = new GitlabProject($project->gitlab_id, $this->client);
        try {
            $gitlabProject->show();
            return true;
        } catch (Throwable $e) {
            if ($e->getMessage() === '404 Project Not Found') {
                return $this->processDeletedProject($project);
            }
            throw $e;
        }
    }

    /**
     * Обработка удаленного проекта GitLab
     * @param AppProject $project
     * @return bool
     * @throws Exception
     */
    public function processDeletedProject(AppProject $project)
    {
        $project->gitlab_id = null;
        $project->save();

        // Очистка gitlab_id у задач и комменариев
        foreach ($project->getIssues()->each(30) as $issue) {
            $issue->gitlab_id = null;
            $issue->save();

            foreach ($issue->getComments()->each(30) as $comment) {
                $comment->gitlab_id = null;
                $comment->save();
            }
        }

        PFApi::syncPlanfixProject($project->planfix_id);

        return true;
    }

    /**
     *
     * Синзронизация проекта
     *
     * @param AppProject $project
     * @param $planfixProject
     * @return AppProject
     * @throws Throwable
     */
    public function syncProject(AppProject $project, $planfixProject)
    {
        if (!$project->gitlab_id) {
            $gitlabProject = GitlabProject::create($this->client, $this->cleanProjectName($project->title), [
                'description' => $project->description,
                'web_url' => $project->gitlab_url,
            ]);
            $project->gitlab_id = $gitlabProject->id;
            $project->save();
        } else {
            $this->checkProject($project); // Check is project deleted
            $gitlabProject = new GitlabProject($project->gitlab_id, $this->client);
            $gitlabProject->update([
                'description' => $project->description,
                'web_url' => $project->gitlab_url,
            ]);
        }

        $this->setProjectHook($gitlabProject, $project);

        $this->projectAddMembers($project, $gitlabProject, $planfixProject);
        PFApi::syncLabels($gitlabProject);

        return $project;
    }

    /**
     * Получение списка участников проекта
     * @param $id
     * @return array
     */
    public function projectAllMembers($id)
    {
        $cache = Yii::$app->cache;
        $data = $cache->getOrSet("project-$id-members", function () use ($id) {
            $this->client->projects()->allMembers($id);
        }, 20);

        $members = array();
        if (!$data || !is_array($data)) {
            return [];
        }

        foreach ($data as $member) {
            $members[] = GitlabUser::fromArray($this->client, $member);
        }

        return $members;
    }

    /**
     * Добавление участников (ПФ) в проект
     * @param $planfixProject
     * @param $gitlabProject
     * @param $type
     * @param $code
     * @throws Exception
     */
    public function addMembers($planfixProject, $gitlabProject, $type, $code)
    {

        $members = [];

        foreach ($this->projectAllMembers($gitlabProject->id) as $member) {
            $members[$member->id] = $member;
        }
        

        if (isset($planfixProject['project'][$type]['users'])) {
            if (isset($planfixProject['project'][$type]['users']['user']['id'])) {
                $planfixProject['project'][$type]['users']['user'] = [ $planfixProject['project'][$type]['users']['user'] ];
            }

            foreach ($planfixProject['project'][$type]['users']['user'] as $worker) {
                $user = PFApi::getUserFromPlanfix($worker['id'], true)->gitlab_id;
                if (isset($members[$user]) && $members[$user]->access_level >= $code) {
                    continue;
                }
                try {
                    if (isset($members[$user])) {
                        $gitlabProject->saveMember($user, $code);
                    } else {
                        $gitlabProject->addMember($user, $code);
                    }
                } catch (Throwable $e) {
                    Yii::$app->impulse->air('app_error', sprintf("Cant add member: %s %s", $worker['id'], $gitlabProject->id));
                }
            }
        }
    }

    /**
     * Добавление участника в проект с проверкой
     * @param $user
     * @param $gitlabProject
     * @param $access
     */
    public function checkAddMember($user, $gitlabProject, $access)
    {
        $members = [];

        foreach ($this->projectAllMembers($gitlabProject->id) as $member) {
            $members[$member->id] = $member;
        }

        if (!isset($members[$user->gitlab_id])) {
            $gitlabProject->addMember($user->gitlab_id, $access);
        } elseif ($members[$user->gitlab_id]->access_level < $access) {
            $gitlabProject->saveMember($user->gitlab_id, $access);
        }
    }

    /**
     * Добавление участников в проект
     * @param $project
     * @param $gitlabProject
     * @param $planfixProject
     * @throws Exception
     */
    public function projectAddMembers($project, $gitlabProject, $planfixProject)
    {
        
        $this->addMembers($planfixProject, $gitlabProject, 'members', 20);
        $this->addMembers($planfixProject, $gitlabProject, 'auditors', 20);
        $this->addMembers($planfixProject, $gitlabProject, 'workers', 30);
        

        $members = [];

        foreach ($this->projectAllMembers($gitlabProject->id) as $member) {
            $members[$member->id] = $member;
        }

        try {
            if ($project->owner && !isset($members[$project->owner->gitlab_id])) {
                $gitlabProject->addMember($project->owner->gitlab_id, 40);
            } elseif ($project->owner && isset($members[$project->owner->gitlab_id]) && $members[$project->owner->gitlab_id]->access_level < 40) {
                $gitlabProject->saveMember($project->owner->gitlab_id, 40);
            }
        } catch (Throwable $e) {
            Yii::$app->impulse->air('error', 'Cant add member ' . $project->owner->id, true);
        }

        try {
            if ($project->lead && !isset($members[$project->lead->gitlab_id])) {
                $gitlabProject->addMember($project->lead->gitlab_id, 40);
            } elseif ($project->lead && !isset($members[$project->lead->gitlab_id]) && $members[$project->lead->gitlab_id]->access_level < 40) {
                $gitlabProject->saveMember($project->owner->gitlab_id, 40);
            }
        } catch (Throwable $e) {
            Yii::$app->impulse->air('error', 'Cant add member ' . $project->lead->id, true);
        }
    }

    /**
     * Синхронизация пользователей
     * @param $user
     * @return mixed
     * @throws Exception
     */
    public function syncUser($user)
    {
        if (!$user->gitlab_id) {
            $users = $this->raw('search', [
                'scope' => 'users',
                'search' => $user->username,
            ]);

            foreach ($users as $searchUser) {
                if ($searchUser->username == $user->username) {
                    $user->gitlab_id = $searchUser->id;
                    $user->save();

                    return $user;
                }
            }

            if (!$this->isAdmin) {
                Yii::$app->impulse->air('app_error', 'Пользователь ' . $user->username . '(' . $user->planfix_id . ') не найден в GitLab. Для создания пользователей требуется аккаунт администратора в GitLab. Проведите ассоциацию вручную, или используйте аккаунт администратора.');
                throw new Exception('Пользователь ' . $user->username . '(' . $user->planfix_id . ') не найден в GitLab');
            }

            $gitlabUser = null;
            try {
                $gitlabUser = GitlabUser::create($this->client, $user->email, bin2hex(random_bytes(30)), [
                    'name' => $user->first_name . ' ' . $user->middle_name . ' ' . $user->last_name,
                    'username' => $user->username,
                ]);
            } catch (Throwable $e) {
                if ($e->getMessage() == '"username" ' . $user->username . ' is a reserved name') {
                    $user->username = 'pf_' . $user->username;
                    $gitlabUser = GitlabUser::create($this->client, $user->email, bin2hex(random_bytes(30)), [
                        'name' => $user->first_name . ' ' . $user->middle_name . ' ' . $user->last_name,
                        'username' => $user->username,
                    ]);
                }
            }
            if (!$gitlabUser) {
                return null;
            }
            $user->gitlab_id = $gitlabUser->id;
            $user->save();
        } else {
            if (!$this->isAdmin) {
                return $user;
            }
            
            $gitlabUser = new GitlabUser($user->gitlab_id, $this->client);
            $gitlabUser->update([
                'name' => $user->first_name . ' ' . $user->middle_name . ' ' . $user->last_name,
            ]);
        }

        return $user;
    }

    /**
     * Отправка запроса к GitLab API
     * @param $method
     * @param array $getParams
     * @param array $postParams
     * @param bool $post
     * @param string $rmethod
     * @return mixed
     */
    public function raw($method, $getParams = [], $postParams = [], $post = false, $rmethod = "GET")
    {
        $opts = [
            "http" => [
                "method" => $post?  "POST" : $rmethod,
                "header" =>
                    "PRIVATE-TOKEN: $this->token\r\n" .
                    ($post? "Content-Type: application/x-www-form-urlencoded\r\n" : ''),
                "content" => ($post || $rmethod != 'GET')? http_build_query($postParams) : null,
                'ignore_errors' => true,
            ]
        ];

        if ($method[0] == '/') {
            $method = substr($method, 1);
        }

        $context = stream_context_create($opts);
        $data = file_get_contents($this->url . '/api/v4/' . $method . '?' . http_build_query($getParams), false, $context);

        $this->logRequest();

        $data = json_decode($data);

        if (!$data) {
            throw new Exception("Request failed: $method " . print_r($getParams, true) . print_r($postParams, true));
        }

        return $data;
    }


    /**
     * Синхронизация Issue
     * @param $issue
     * @param $planfixTask
     * @param null $user_id
     * @throws Exception
     */
    public function syncIssue($issue, $planfixTask, $user_id = null)
    {
        $project = AppProject::findOne([
            'planfix_id' => $planfixTask['task']['project']['id'],
        ]);

        $markdown = new HtmlConverter();
        $text = $markdown->convert($issue->description);

        if (!isset(PFApi::getPlanfixStatusList()[$issue->status])) {
            $labels = PFApi::getPlanfixStatus($issue->status, $project);
        } else {
            $labels = PFApi::getPlanfixStatusList()[$issue->status]['name'];
        }

        $statusMapping = CoreStatus::find()->where([
            'planfix_id' => $issue->status,
            'project_id' => $issue->project->id
        ])->one();

        if ($statusMapping) {
            $labels = $statusMapping->label;
        }
        

        $issueData = [
            'title'       => $issue->title,
            'description' => $text,
            'assignee_id' => $issue->assignee? $issue->assignee->gitlab_id : null,
            'sudo'        => $issue->creator->gitlab_id,
            'due_date'    => $issue->due_date,
            'labels'      => $labels,
        ];

        if (!$this->isAdmin) {
            unset($issueData['sudo']);
        }

        $issue->description_gitlab = $text;
        $issue->save();

        if ($issue->labels != 'Завершенная' && $labels == 'Завершенная') {
            $issueData['state_event'] = 'close';
        }

        if ($issue->labels == 'Завершенная' && $issueData['labels'] != 'Завершенная') {
            $issueData['state_event'] = 'reopen';
        }

        

        $gitlabProject = new GitlabProject($project->gitlab_id, $this->client);

        $members = [];

        foreach ($gitlabProject->members() as $member) {
            $members[$member->id] = $member;
        }

        if (!isset($members[$issue->creator->gitlab_id])) {
            if (!$issue->creator->gitlab_id) {
                $this->syncUser($issue->creator)->gitlab_id;
            }
            $this->checkAddMember($issue->creator, $gitlabProject, 20);
        }

        if ($issue->assignee && !isset($members[$issue->assignee->gitlab_id])) {
            if (!$issue->assignee->gitlab_id) {
                $this->syncUser($issue->assignee)->gitlab_id;
            }
            $this->checkAddMember($issue->assignee, $gitlabProject, 30);
        }

        if (!$issue->gitlab_id) {
            if (!$project) {
                PFApi::syncPlanfixProject($planfixTask['task']['project']['id']);

                $project = AppProject::findOne([
                    'planfix_id' => $planfixTask['task']['project']['id'],
                ]);
                if (!$project) {
                    throw new Exception("No project found!");
                }
            }
           
            $gitlabIssue = $this->raw('projects/' . $project->gitlab_id . '/issues', ($this->isAdmin? [
                'sudo' => $issue->creator->gitlab_id,
            ] : []), $issueData, true);

            if (!property_exists($gitlabIssue, 'iid')) {
                throw new Exception('Issue iid doesnt exists: ' . print_r($gitlabIssue, true) . print_r($project->gitlab_id, true) . print_r($issueData, true));
            }

            $issue->project_id = $project->id;
            $issue->gitlab_id = $gitlabIssue->iid;

            $issue->save();
        } else {
            $this->raw('projects/' . $project->gitlab_id . '/issues/' . $issue->gitlab_id, ($this->isAdmin? [
                'sudo' => $user_id? User::findOne($user_id)->gitlab_id : $issue->creator->gitlab_id,
            ] : []), $issueData, false, 'PUT');
        }

        $this->syncIssueComments($issue);
    }


    /**
     * Синхронизация комментариев
     * @param Issue $issue
     * @throws Exception
     */
    public function syncIssueComments(Issue $issue)
    {
        $commentsToAdd = IssueComment::find()->where([
            'or',
            ['gitlab_id' => null],
            ['issue_id' => $issue->id],
        ])->all();

        foreach ($commentsToAdd as $comment) {
            $this->syncIssueComment($issue, $comment);
        }
    }

    /**
     * @param $user
     * @param $gitlabProject
     */
    public function checkAddMemberShip($user, $gitlabProject)
    {
        try {
            return $this->checkAddMember($user, $gitlabProject, 20);
        } catch (\Throwable $e) {
            // TODO Fix
        }
    }

    /**
     * Синхронизация комментария
     * @param $issue
     * @param $comment
     * @throws Exception
     */
    public function syncIssueComment(Issue $issue, IssueComment $comment)
    {
        $issue->refresh();
        $issue->project->refresh();
        $gitlabProject = new GitlabProject($issue->project->gitlab_id, $this->client);
        $gitlabIssue = new GitlabIssue($gitlabProject, $issue->gitlab_id, $this->client);

        if (!$comment->gitlab_id) {
            if ($comment->getGitlabText()) {
                $this->checkAddMemberShip($comment->creator, $gitlabProject);

                $note = $this->raw("/projects/$gitlabProject->id/issues/$gitlabIssue->iid/notes", ($this->isAdmin? [
                    'sudo' => $comment->creator->gitlab_id,
                    'body' => $comment->getGitlabText(),
                    'created_at' => date('c', strtotime($comment->created_at)),
                ] : [
                    'created_at' => date('c', strtotime($comment->created_at)),
                    'body' => $comment->getGitlabText()
                ]), [], true, 'POST');

                if (!property_exists($note, 'id')) {
                    throw new Exception(print_r($note, true));
                }

                $comment->gitlab_id = $note->id;
                $comment->text_gitlab = $comment->getGitlabText();
                $comment->save();
            }
        } else {
            $this->raw("/projects/$gitlabProject->id/issues/$gitlabIssue->iid/notes/$comment->gitlab_id", ($this->isAdmin? [
                'sudo' => $comment->creator->gitlab_id,
                'body' => $comment->getGitlabText()
            ] : [
                'body' => $comment->getGitlabText()
            ]), [], false, 'PUT');

            $comment->text_gitlab = $comment->getGitlabText();
            $comment->save();
        }
    }

    /**
     * Добавление вебхука к проекту
     *
     * @param $gitlabProject
     * @param $dbProject
     *
     * @return bool
     * @throws Exception
     */
    public function setProjectHook($gitlabProject, $dbProject)
    {

        if ($this->noHooks) {
            return true;
        }

        $url = Yii::getAlias('@frontendUrl') . '/gitlab/webhook';

        $hooks = $gitlabProject->hooks();

        $hookFound = false;
        foreach ($hooks as $hook) {
            if ($hook->url == $url) {
                $hookFound = true;
            }
        }

        if (!$hookFound) {
            $token = bin2hex(random_bytes(20));

            $dbProject->token = $token;
            $dbProject->save();

            $gitlabProject->addHook(
                $url,
                [
                    'push_events'                => true,
                    'issues_events'              => true,
                    'confidential_issues_events' => true,
                    'merge_requests_events'      => true,
                    'tag_push_events'            => true,
                    'note_events'                => true,
                    'job_events'                 => true,
                    'pipeline_events'            => true,
                    'wiki_page_events'           => true,
                    'token'                      => $token,
                    'enable_ssl_verification'    => true,
                ]
            );

            return true;
        }

        return false;
    }

    /**
     * @param $issueId
     * @param $projectId
     * @param bool $notes
     * @param bool $hasChanges
     * @return bool
     * @throws Exception
     */
    public function importIssue($issueId, $projectId, $notes = false, $hasChanges = false)
    {

        $project = AppProject::findOne([
            'id' => $projectId,
        ]);

        if (!$project) {
            return false;
        }

        $issue = Issue::findOne(['gitlab_id' => "$issueId", 'project_id' => $project->id]);
        // $hasChanges = false;
        if (!$issue) {
            $issue = new Issue();
            $issue->gitlab_id = $issueId;
            $issue->project_id = $project->id;

            $notes = true;
            $hasChanges = true;
        }

        $gitlabIssue = $this->client->issues()->show($project->gitlab_id, $issueId);

        if (!$gitlabIssue) {
            return false;
        }
        
        $title = $gitlabIssue['title'];
        $created_at = date('Y-m-d H:i:s', strtotime($gitlabIssue['created_at']));

        $description_gitlab = $gitlabIssue['description'];
        $due_date = $gitlabIssue['due_date']? (date('Y-m-d H:i:s', strtotime($gitlabIssue['due_date']))) : null;
        
        $labels = implode(',', $gitlabIssue['labels']);

        $assignee_id = null;
        if (isset($gitlabIssue['assignees'][0])) {
            $assignee_id = $this->getUser($gitlabIssue['assignees'][0]['id'])->id;
        } else {
            $assignee_id = $this->getUser($gitlabIssue['author']['id'])->id;
        }

        $creator_id = $this->getUser($gitlabIssue['author']['id'])->id;

        if ($title != $issue->title || $created_at != $issue->created_at || $description_gitlab != $issue->description_gitlab ||
            $due_date != $issue->due_date || $labels != $issue->labels || $assignee_id != $issue->assignee_id || $creator_id != $issue->creator_id) {
            $hasChanges = true;
        }

        $issue->title = $title;
        $issue->created_at = $created_at;
        $issue->description_gitlab = $description_gitlab;
        $issue->due_date = $due_date;
        $issue->labels = $labels;
        $issue->assignee_id = $assignee_id;
        $issue->creator_id = $creator_id;

        $save = $issue->save();

        $user = $issue->creator;
        if ($user && !$user->init_issue_id) {
            $user->init_issue_id = $issue->id;
            $user->save();
        }

        $user = $issue->assignee;
        if ($user && !$user->init_issue_id) {
            $user->init_issue_id = $issue->id;
            $user->save();
        }

        if (!$save) {
            throw new Exception(print_r($issue->getErrors(), true));
        }

        if ($hasChanges) {
            PFApi::syncIssueGitlab($issue);
        }

        if ($notes) {
            $this->importNotes($issue, $project);
        }

        return true;
    }

    /**
     * @param $gitlabId
     * @return User|array|ActiveRecord
     * @throws Exception
     */
    public function getUser($gitlabId)
    {
        $user = User::findOne(['gitlab_id' => $gitlabId]);

        if ($user) {
            return $user;
        }

        $gitlabUser = $this->client->users()->show($gitlabId);

        $user = User::find()->where(['email' => $gitlabUser['email'], 'gitlab_id' => null])->one();

        if ($user) {
            $user->gitlab_id = $gitlabId;
            $user->save();

            return $user;
        }

        $user = new User();

        $user->email = $gitlabUser['email'];
        $user->username = str_replace(' ', '-', $gitlabUser['username']);

        // $user->access_token = '';

        $user->gitlab_id = $gitlabId;


        $nameSurname         = explode(' ', $gitlabUser['name']);
        $user->first_name = $nameSurname[0];

        if (count($nameSurname) > 1) {
            $user->last_name = $nameSurname[1];
        }
        
        $save = $user->save();

        if (!$save) {
            throw new Exception(print_r($user->getErrors(), true));
        }

        // PF - сотрудники
        $pUsers = ApiHelper::getUsers();
        
        foreach ($pUsers as $planfixUser) {
            if ($planfixUser['email'] === $user->email) {
                $user->planfix_id = $planfixUser['id'];
                $user->save();
                return $user;
            }

            if ($planfixUser['login'] === $user->username) {
                $user->planfix_id = $planfixUser['id'];
                $user->save();
                return $user;
            }
        }

        return $user;
    }

    /**
     * Синхронизация отсутствующих задач / комментариев
     */
    public function syncMissing()
    {
        $missing = Issue::find()->where([
            'planfix_id' => null
        ])->all();

        foreach ($missing as $issue) {
            try {
                $this->importIssue($issue->gitlab_id, $issue->project->id, true, true);
            } catch (Exception $e) {
            }
        }

        $missingNotes = IssueComment::find()->where([
            'planfix_id' => null
        ])->all();

        foreach ($missingNotes as $note) {
            try {
                $this->importNote($note->gitlab_id, $note->issue->gitlab_id, $note->issue->project->id, true);
            } catch (Exception $e) {
            }
        }
    }

    /**
     * Импорт коммнтариев
     * @param $issue
     * @param $project
     * @throws Exception
     */
    public function importNotes($issue, $project)
    {
        $comments = $this->client->issues()->showComments($project->gitlab_id, $issue->gitlab_id);

        foreach ($comments as $comment) {
            $this->importNote($comment['id'], $issue->gitlab_id, $project->id);
        }
    }


    /**
     * Импорт комменатрия
     * @param $id
     * @param $issue_id
     * @param $project_id
     * @param bool $hasChanges
     * @throws Exception
     */
    public function importNote($id, $issue_id, $project_id, $hasChanges = false)
    {
        $note = IssueComment::findOne([
            'gitlab_id' => $id
        ]);
    
        $issue = Issue::findOne([
            'project_id' => $project_id,
            'gitlab_id' => $issue_id
        ]);

        $gitlabNote = $this->raw("/projects/{$issue->project->gitlab_id}/issues/$issue->gitlab_id/notes/$id");

        if (!$note) {
            $note = new IssueComment();
            $note->gitlab_id = $id;
            $note->issue_id = $issue->id;
            $note->text = '';
            $note->pf_time = '';
            $note->pf_time_old = '';
            $note->origin = 'gitlab';

            $creator = User::findOne([
                'gitlab_id' => $gitlabNote->author->id,
            ]);

            if (!$creator) {
                $creator = $this->getUser($gitlabNote->author->id);
            }
            $note->creator_id = $creator->id;
            $note->created_at = date('Y-m-d H:i:s', strtotime($gitlabNote->created_at));

            $hasChanges = true;
        }
        
        if ($note->text_gitlab != $gitlabNote->body) {
            $hasChanges = true;
        }
        $note->text_gitlab = $gitlabNote->body;
        $note->is_log = $gitlabNote->system? 1: 0;

        if (trim($note->text_gitlab) == 'Добавлено время из Планфикс') {
            $note->is_log = 1;
        }

        if (trim($note->text_gitlab) == 'Удалено время из Planfix') {
            $note->is_log = 1;
        }

        if (strpos($note->text_gitlab, 'changed time estimate') !== false && $note->is_log) {
            $note->is_log = 0;
        }

        if (strpos($note->text_gitlab, 'subtracted') !== false && $note->is_log) {
            $note->is_log = 0;
            $note->text_gitlab = 'Изменено время в GitLab: ' . $note->text_gitlab;
        }

        if (strpos($note->text_gitlab, 'removed time spent') !== false && $note->is_log) {
            $note->is_log = 0;
            $note->text_gitlab = 'Удалено время в GitLab: ' . $note->text_gitlab;
        }



        $note->files = '';

        $usersToNotify = [];
        // Ищем упомянутых в комментарии пользователей
        preg_match_all('/@[A-z0-9-]{1,}/', $note->text_gitlab, $matches);

        if (isset($matches[0])) {
            foreach ($matches[0] as $username) {
                $gitlabUser = $this->raw('users', ['username' => substr($username, 1)]);

                if (isset($gitlabUser[0])) {
                    $gitlabUser = $gitlabUser[0];
                    $user = User::findOne(['gitlab_id' => $gitlabUser->id]);
                } else {
                    $user = null;
                    Yii::$app->impulse->air('app_error', 'User with username ' . substr($username, 1) . ' not found in Gitlab');
                }

              
                if ($user && $user->planfix_id) {// && $model->creator_id != $user->id) {
                    $usersToNotify[] = $user->id; // Добавляем в список, если нашли
                }
            }
        }
        $notifyList = implode(',', $usersToNotify);
        if ($note->notify_list != $notifyList) {
            $hasChanges = true;
        }

        $note->notify_list = $notifyList;

        $save = $note->save();

        $user = $note->creator;
        if ($user && !$user->init_issue_id) {
            $user->init_issue_id = $note->issue_id;
            $user->save();
        }

        if (!$save) {
            throw new Exception(print_r($note->getErrors(), true));
        }

        if ($hasChanges) {
            PFApi::syncNoteGitlab($note);
        }
    }

    /**
     * Получение времени из комментария
     * @param $comment
     * @return array|null
     * @throws Exception
     */
    public function getCommentTime($comment)
    {
        $hasTimeAdded = preg_match(
            '/^(added )([a-z0-9 ]{1,})( of time spent)( at )?([0-9-]{10,})?/',
            $comment->text_gitlab,
            $matches
        );

        if ($hasTimeAdded) {
            $amount = $this->parseIssueTime($matches[2]);

            if (!$amount) {
                throw new Exception("Cant parse time " . $comment->text, 1);
            }

            return [
                'amount' => $amount,
                'date' => isset($matches[5]) ? $matches[5] : $comment->created_at,
            ];
        }

        return null;
    }



    /**
     * @param $gitlabId
     *
     * @return bool|User|null
     * @throws Exception
     */
    public function importUser($gitlabId)
    {
        $user = User::findOne(['gitlab_id' => $gitlabId]);

        $gitlabUser = $this->client->api('users')->show($gitlabId);

        if (!$user) {
            $user = new User();
            $user->username = str_replace(' ', '-', $gitlabUser['username']);
            $user->email = $gitlabUser['email'];
            $user->gitlab_id = $gitlabId;
        }
       
        $nameSurname         = explode(' ', $gitlabUser['name']);
        $user->first_name = $nameSurname[0];

        if (count($nameSurname) > 1) {
            $user->last_name = $nameSurname[1];
        }

        $save = $user->save();

        if (!$user || !$save) {
            return false;
        }

        return $user;
    }


    /**
     * Перевод времени из формата Gitlab в секунды
     *
     * @param $timeText - Время из комментария (1y 1mo 2w 1d 1h 1m 1s)
     *
     * @return int - Время в секундах
     */
    public function parseIssueTime($timeText)
    {
        //
        $amount = 0;

        if (preg_match('/([0-9]{1,})(s)/', $timeText, $matches)) {
            $amount += $matches[1];
        }

        if (preg_match('/([0-9]{1,})(m)/', $timeText, $matches)) {
            $amount += $matches[1] * 60;
        }

        if (preg_match('/([0-9]{1,})(h)/', $timeText, $matches)) {
            $amount += $matches[1] * 60 * 60;
        }

        if (preg_match('/([0-9]{1,})(d)/', $timeText, $matches)) {
            $amount += $matches[1] * 8 * 60 * 60;
        }

        if (preg_match('/([0-9]{1,})(w)/', $timeText, $matches)) {
            $amount += $matches[1] * 5 * 8 * 60 * 60;
        }

        if (preg_match('/([0-9]{1,})(mo)/', $timeText, $matches)) {
            $amount += $matches[1] * 4 * 5 * 8 * 60 * 60;
        }

        if (preg_match('/([0-9]{1,})(y)/', $timeText, $matches)) {
            $amount += $matches[1] * 12 * 4 * 5 * 8 * 60 * 60;
        }

        return $amount;
    }






    /**
     * @param AppProject $model
     *
     * @return GitlabProject
     */
    public function loadProject(AppProject $model)
    {
        return new GitlabProject($model->gitlab_id, $this->client);
    }

    /**
     * Рендеринг gitlab markdown
     * @param $text
     * @param $project
     * @return mixed
     */
    public function renderMarkDown($text, $project)
    {

        $gitlabProject = new GitlabProject($project, $this->client);
        $project = $gitlabProject->show();

        $data = $this->raw('markdown', [], [
            'text' => $text,
            'gfm' => false,
            'project' => $project->path_with_namespace
        ], true, "POST");


        return $data->html;
    }
}
