<?php

namespace app\planfix\classes;

use app\core\helpers\TimeSyncHelper;
use Yii;
use app\core\models\AppUser as User;
use app\core\models\Project;
use app\planfix\helpers\ApiHelper;
use app\core\models\Issue;
use Exception;
use app\core\models\IssueComment;
use Gitlab\Model\Project as GitlabProject;
use app\core\models\CoreStatus;
use app\core\models\Time;
use Throwable;

/**
 * Class PFApi
 * @package app\planfix\classes
 */
class PFApi
{
    /**
     * @param $id
     * @return User|null
     */
    public static function findUser($id)
    {
        try {
            $planfixUser = Yii::$app->planfix->api('user.get', [
                'user' => [
                    'id' => (int) $id
                ]
            ]);
        } catch (Exception $e) {
            try {
                $planfixUser = Yii::$app->planfix->api('contact.get', [
                    'contact' => [
                        'id' => (int) $id
                    ]
                ]);

                $planfixUser['user'] = $planfixUser['contact'];
                unset($planfixUser['contact']);
            } catch (Exception $e) {
                return null;
            }
        }

        return $planfixUser;
    }

    /**
     * @param $planfix_id
     * @param bool $full
     * @param null $gitlab_id
     * @param null $issueId
     * @param null $projectId
     * @return User|int|null
     * @throws Exception
     */
    public static function getUserFromPlanfix($planfix_id, $full = false, $gitlab_id = null, $issueId = null, $projectId = null)
    {

        if (!$planfix_id) {
            return null;
        }

        $user = User::findOne([
            'planfix_id' => $planfix_id,
        ]);

        if (!$user) {
            $planfixUser = static::findUser($planfix_id);

            if (!$planfixUser) {
                return null;
            }

            $user = new User();
            $user->username = (isset($planfixUser['user']['login']) && !is_array($planfixUser['user']['login']))? $planfixUser['user']['login']: explode('@', $planfixUser['user']['email'])[0];

            if (User::find()->where(['username' => $user->username ])->count()) {
                if ((User::find()->where(['username' => 'pf_' . $user->username])->count())) {
                    $user->username = 'pf_' . uniqid() . '_' . $user->username;
                } else {
                    $user->username = 'pf_' . $user->username;
                }
            }

            $user->gitlab_id = $gitlab_id;
            
            if (isset($planfixUser['user']['login']) && is_array($planfixUser['user']['login'])) {
                unset($planfixUser['user']['login']);
            }

            $user->email = isset($planfixUser['user']['login'])? ($planfixUser['user']['login'] . '@' . Yii::$app->planfix->account . '.planfix.ru') : $planfixUser['user']['email'];
            $user->planfix_id = $planfix_id;
            $user->init_project_id = $projectId;
            $user->init_issue_id = $issueId;

            $user->first_name = is_array($planfixUser['user']['name']) ? '' : $planfixUser['user']['name'];
            $user->last_name = is_array($planfixUser['user']['lastName']) ? '' : $planfixUser['user']['lastName'];

            $save = $user->save();
            if (!$save) {
                throw new Exception(print_r($user->getErrors(), true));
            }
        } else {
            $planfixUser = static::findUser($planfix_id);

            if (!$planfixUser) {
                return null;
            }
            
            $user->first_name = is_array($planfixUser['user']['name']) ? '' : $planfixUser['user']['name'];
            $user->last_name = is_array($planfixUser['user']['lastName']) ? '' : $planfixUser['user']['lastName'];
            $user->save();
        }
        
        Yii::$app->gitlab->syncUser($user);

        if ($full) {
            return $user;
        }

        return $user->id;
    }

    /**
     * @param Project $project
     * @return bool
     * @throws Throwable
     */
    public static function checkProject(Project $project)
    {
        $planfix = Yii::$app->planfix;

        try {
            $planfix->api('project.get', [
                'project' => [
                    'id' => $project->planfix_id,
                ]
            ]);

            return true;
        } catch (Throwable $e) {
            if ($e->getMessage() === 'Запрошенный проект не существует -> project') {
                return static::processDeletedProject($project);
            }
            throw $e;
        }
    }

    /**
     * @param Project $project
     * @return bool
     */
    public static function processDeletedProject(Project $project)
    {
        $project->planfix_id = null;
        $project->save();

        foreach ($project->getIssues()->each(30) as $issue) {
            $issue->planfix_id = null;
            $issue->save();

            foreach ($issue->getComments()->each(30) as $comment) {
                $comment->gitlab_id = null;
                $comment->save();
            }
        }

        // Create Project
        //
        static::syncProjectFromModel($project);

        return true;
    }

    /**
     * @param Project $project
     */
    public static function syncProjectFromModel(Project $project)
    {
        if (!$project->planfix_id) {
            $response = Yii::$app->planfix->api('project.add', [
                'project' => [
                    'title' => $project->title,
                    'description' => $project->description,
                    'owner' => [
                        'id' => $project->owner? $project->owner->planfix_id : null,
                    ]
                ]
            ], false, true);

            $project->planfix_id = $response['project']['id'];
            $project->save();
        }
    }

    /**
     * @param $planfixId
     * @param null $gitlabId
     * @return mixed
     * @throws Exception
     */
    public static function syncPlanfixProject($planfixId, $gitlabId = null)
    {

        $planfix = Yii::$app->planfix;
        $planfixProject = $planfix->api('project.get', [
            'project' => [
                'id' => $planfixId,
            ]
        ]);

        // return $planfixProject;

        $project = Project::findOne([
            'planfix_id' => $planfixId,
        ]);

        if (!$project) {
            $project = new Project();
            $project->planfix_id = $planfixProject['project']['id'];
            $project->gitlab_id = $gitlabId;
        }

        $project->title = $planfixProject['project']['title'];
        $project->description = $planfixProject['project']['description'];
        
        if (is_array($project->description)) {
            $project->description = '';
        }

        $project->owner_id = PFApi::getUserFromPlanfix($planfixProject['project']['owner']['id']);
        $project->lead_developer_id = PFApi::getUserFromPlanfix(
            ApiHelper::findCustomDataProperty(Yii::$app->planfix->pr_field_lead_dev, $planfixProject['project'])['value']
        );
        
        $save = $project->save();

        // User came from this project
        $user = $project->owner;
        if (!$user->init_project_id) {
            $user->init_project_id = $project->id;
            $user->save();
        }

        $user = $project->lead;
        if ($user && !$user->init_project_id) {
            $user->init_project_id = $project->id;
            $user->save();
        }

        if (!$save) {
            throw new Exception(print_r($project->getErrors(), true));
        }

        return Yii::$app->gitlab->syncProject($project, $planfixProject);
    }

    /**
     * @param bool $byNames
     * @return array
     * @throws Exception
     */
    public static function getPlanfixStatusList($byNames = false)
    {
        $processId = null;
        $taskProcesses = Yii::$app->planfix->api('taskStatus.getSetList');

        $processId = Yii::$app->planfix->process_id;

        if (!$processId) {
            foreach ($taskProcesses['taskStatusSets']['taskStatusSet'] as $set) {
                if ($set['name'] == 'Разработка') {
                    $processId = $set['id'];
                }
            }
        }

        if (!$processId) {
            throw new Exception("Процесс разработки не найден");
        }

        $statuses = Yii::$app->planfix->api('taskStatus.getListOfSet', [
            'taskStatusSet' => [
                'id' => $processId,
            ]
        ]);

        $res = [];

        foreach ($statuses['taskStatuses']['taskStatus'] as $status) {
            if (!$byNames) {
                $res[$status['id']] = $status;
            } else {
                $res[$status['name']] = $status;
            }
        }

        return $res;
    }

    /**
     * @param $statusId
     * @param null $project
     * @param bool $import
     * @param bool $returnId
     * @return mixed
     * @throws Exception
     */
    public static function getPlanfixStatus($statusId, $project = null, $import = true, $returnId = false)
    {
        $taskProcesses = Yii::$app->planfix->api('taskStatus.getSetList');

        foreach ($taskProcesses['taskStatusSets']['taskStatusSet'] as $set) {
            $statuses = Yii::$app->planfix->api('taskStatus.getListOfSet', [
                'taskStatusSet' => [
                    'id' => $set['id'],
                ]
            ]);

            foreach ($statuses['taskStatuses']['taskStatus'] as $status) {
                if ($status['id'] == $statusId) {
                    if ($import) {
                        $gitlabProject = new GitlabProject($project->gitlab_id, Yii::$app->gitlab->client);
                        static::syncLabels($gitlabProject, $set['id']);
                    }
                    if ($returnId) {
                        return $status['id'];
                    }

                    return $status['name'];
                }
            }
        }

        return true;
    }


    /**
     * @param $gitlabProject
     * @param null $setId
     * @throws Exception
     */
    public static function syncLabels($gitlabProject, $setId = null)
    {
        $processId = null;
        $taskProcesses = Yii::$app->planfix->api('taskStatus.getSetList');

        if (!$setId) {
            $sets = $taskProcesses['taskStatusSets']['taskStatusSet'];
            if (!isset($sets[0])) {
                $sets = [ $sets ];
            }

            foreach ($sets as $set) {
                if ($set['name'] == 'Разработка') {
                    $processId = $set['id'];
                }
            }

        } else {
            $processId = $setId;
        }

        if (!$processId) {
            throw new Exception("Процесс разработки не найден");
        }

        $statuses = Yii::$app->planfix->api('taskStatus.getListOfSet', [
            'taskStatusSet' => [
                'id' => $processId,
            ]
        ]);

        $labelByName = [];
        $labels = $gitlabProject->labels();

        foreach ($labels as $label) {
            $labelByName[$label->name] = $label;
        }

        foreach ($statuses['taskStatuses']['taskStatus'] as $status) {
            if (!isset($labelByName[$status['name']])) {
                try {
                    $gitlabProject->addLabel($status['name'], $status['color']);
                } catch (Exception $e) {
                }
            }
        }
    }

    /**
     * @param $issue
     * @return bool
     */
    public static function checkIssueTime($issue)
    {
        $issueTime = Time::find()->where(['issue_id' => $issue->id])->all();

        if (!count($issueTime)) {
            return true;
        }

        $keys = [];
        $ids = [];
        $idsFound = [];

        foreach ($issueTime as $time) {
            $keys[] = ['key' => $time->planfix_id];
            $ids[] = $time->planfix_id;
        }

        $data = Yii::$app->planfix->api('analitic.getData', [
            'analiticKeys' => $keys
        ]);

        if (!isset($data['analiticDatas']) || !isset($data['analiticDatas']['analiticData'])) {
            $data = [ 'analiticDatas' => [ 'analiticData' => []]];
        }
        $pfTime = $data['analiticDatas']['analiticData'];

        if (isset($pfTime['key'])) {
            $pfTime = [ $pfTime ];
        }

        foreach ($pfTime as $analitic) {
            $idsFound[] = intval($analitic['key']);
        }

        foreach ($ids as $key) {
            if (in_array($key, $idsFound)) {
                continue;
            }

            TimeSyncHelper::deletePlanfixTime($key);
        }

        return true;
    }

    /**
     * @param $planfixId
     * @param bool $syncComments
     * @param bool $isChanges
     * @throws Exception
     */
    public static function syncIssuePlanfix($planfixId, $syncComments = true, $isChanges = false)
    {

        $planfixTask = Yii::$app->planfix->api('task.get', [
            'task' => [
                'general' => $planfixId,
            ]
        ]);


        // die(print_r($planfixTask, true));

        $issue = Issue::findOne([
            'planfix_id' => $planfixId
        ]);

        // $isChanges = false;

        if (!$issue) {
            $issue = new Issue();
            $issue->planfix_id = $planfixId;

            $taskProject = Project::find()->where([
                'planfix_id' => $planfixTask['task']['project']['id']
            ])->one();

            if ($taskProject) {
                $issue->project_id = $taskProject->id;
            }


            $syncComments = true;
            $isChanges = true;
        }

        $description = is_array($planfixTask['task']['description'])? '' : $planfixTask['task']['description'];
        $due_date = (isset($planfixTask['task']['endTime']) && is_string($planfixTask['task']['endTime']))? date('Y-m-d H:i:s', strtotime($planfixTask['task']['endTime'])) : null;
        $title = $planfixTask['task']['title'];
        $status = $planfixTask['task']['status'];

        $creatorId = PFApi::getUserFromPlanfix($planfixTask['task']['owner']['id']);

        $assignee = null;
        if (isset($planfixTask['task']['workers']['users'])) {
            if (isset($planfixTask['task']['workers']['users']['user']['id'])) {
                $planfixTask['task']['workers']['users']['user'] = [ $planfixTask['task']['workers']['users']['user'] ];
            }
            if (count($planfixTask['task']['workers']['users']['user'])) {
                $assignee = static::getUserFromPlanfix($planfixTask['task']['workers']['users']['user'][0]['id']);
            }
        }

        if ($title != $issue->title || strip_tags($description) != strip_tags($issue->description) || $due_date != $issue->due_date || $creatorId != $issue->creator_id || $status != $issue->status || $assignee != $issue->assignee_id) {
            $isChanges = true;
        }

        $issue->title = $title;
        $issue->description = $description;
        $issue->due_date = $due_date;
        $issue->status = $status;
            
        $issue->creator_id = $creatorId;
        $issue->assignee_id = $assignee;

        $save = $issue->save();

        if (!$save) {
            throw new Exception(print_r($issue->getErrors(), true));
        }


        $user = $issue->creator;
        if (!$user->init_issue_id) {
            $user->init_issue_id = $issue->id;
            $user->save();
        }

        Yii::$app->gitlab->checkProject($issue->project);
        $issue->project->refresh();

        if ($isChanges) {
            // throw new \Exception("HAs Changes: " . print_r([strip_tags($description), strip_tags($od)], true));
            Yii::$app->gitlab->syncIssue($issue, $planfixTask);
        }

        if ($syncComments) {
            static::syncIssuePlanfixActions($issue);
            // Yii::$app->gitlab->syncIssueComments($issue);
        }

        static::checkIssueTime($issue);
    }

    /**
     * @param $issue
     * @throws Exception
     */
    public static function syncIssuePlanfixActions($issue)
    {
        $actions = ApiHelper::paginate('action.getList', [
            'task' => [
                'general' => $issue->planfix_id,
            ]
        ], 'actions', 'action');
        
        if (isset($actions['id'])) {
            $actions = [$actions];
        }
        foreach ($actions as $action) {
            if (!in_array($action['type'], ['COMMENT', 'ACTION'])) {
                continue;
            }

            static::syncIssuePlanfixAction($action, $issue);
        }
    }

    /**
     * @param $action
     * @param $issue
     * @param null $gitlab_id
     * @return null
     * @throws Exception
     */
    public static function getAnaliticFromAction($action, $issue, $gitlab_id = null)
    {
        $analitics = [];
        $timeAnalytics = [];

        if (!isset($action['analitics']) || !isset($action['analitics']['analitic'])) {
            return null;
        }

        if (isset($action['analitics']['analitic']['id'])) {
            $analitics = [ $action['analitics']['analitic'] ];
        }

        foreach ($analitics as $record) {
            if ($record['id'] != Yii::$app->planfix->time_analytic_id) {
                continue;
            }
            $timeAnalytics[] = [ 'key' => $record['key'] ];
        }

        if (!count($timeAnalytics)) {
            return null;
        }

        $data = Yii::$app->planfix->api('analitic.getData', [
            'analiticKeys' => $timeAnalytics
        ]);


        if (!isset($data['analiticDatas']) || !isset($data['analiticDatas']['analiticData'])) {
            return null;
        }
        $pfTime = $data['analiticDatas']['analiticData'];

        if (isset($pfTime['key'])) {
            $pfTime = [ $pfTime ];
        }

        $time = [];
        $timeString = '';

        foreach ($pfTime as $record) {
            $entry = ['key' => $record['key']];

            foreach ($record['itemData'] as $itemData) {
                if ($itemData['id'] == Yii::$app->planfix->time_analytic_date_id) {
                    $entry['date'] = $itemData['value'];
                    $timeString .= $itemData['value'];
                }

                if ($itemData['id'] == Yii::$app->planfix->time_analytic_time_id) {
                    $entry['time'] = $itemData['value'];
                    $timeString .= $itemData['value'];
                }

                if ($itemData['id'] == Yii::$app->planfix->time_analytic_user_id) {
                    $entry['user'] = $itemData['valueId'];
                }
            }
            $time[] = $entry;
        }

        foreach ($time as $timeRecord) {
            $dbTime = Time::find()->where(['planfix_id' => $timeRecord['key']])->one();
            $hasChanges = false;

            if (!$dbTime) {
                $dbTime = new Time();
                $dbTime->planfix_id = $timeRecord['key'];
                $hasChanges = true;
            }

            $dbTimeUser_id = static::getUserFromPlanfix($timeRecord['user']);
            $timeArr = explode(' - ', $timeRecord['time']);
            $dbTimeTime = strtotime($timeArr[1]) - strtotime($timeArr[0]);
            $dbTimeDate = date('Y-m-d H:i', strtotime($timeRecord['date'] . ' ' . $timeArr[0]));

            $oldDbTime = null;


            if ($dbTimeUser_id != $dbTime->user_id || $dbTimeTime != $dbTime->time || strtotime($dbTime->date) != strtotime($dbTimeDate)) {
                $hasChanges = true;

                if ($dbTime->id) {
                    $oldDbTime = clone $dbTime;
                }
            }

            $dbTime->user_id = $dbTimeUser_id;
            $dbTime->time = $dbTimeTime;
            $dbTime->date = $dbTimeDate;
            $dbTime->gitlab_id = $gitlab_id;
            $dbTime->issue_id = $issue->id;
            $dbTime->action_id = $action['id'];
            $dbTime->save();

            $user = $dbTime->user;
            if (!$user->init_issue_id) {
                $user->init_issue_id = $dbTime->issue_id;
                $user->save();
            }

            if ($hasChanges && !$gitlab_id) {
                TimeSyncHelper::syncTimePlanfix($dbTime, $issue, $oldDbTime);
            }
        }

        return true;
    }

    /**
     * @param $action
     * @param $issue
     * @throws Exception
     */
    public static function syncIssuePlanfixAction($action, $issue)
    {
        $dbAction = IssueComment::findOne([
            'planfix_id' => $action['id'],
        ]);

        $hasChanges = false;

        if (!$dbAction) {
            $dbAction = new IssueComment();
            $dbAction->planfix_id = $action['id'];
            $dbAction->text_gitlab = '';
            $dbAction->pf_time = '';
            $dbAction->pf_time_old = '';
            $dbAction->origin = 'planfix';

            $hasChanges = true;
        }

        $text = is_array($action['description'])? '' : $action['description'];
        if (strip_tags($dbAction->text) != strip_tags($text)) {
            $hasChanges = true;
        }

        $dbAction->text = $text;
        $dbAction->issue_id = $issue->id;

        $dbAction->creator_id = static::getUserFromPlanfix($action['owner']['id']);
        $dbAction->is_log = 0;
        $dbAction->planfix_id = $action['id'];
        $dbAction->created_at = date('Y-m-d H:i:s', strtotime($action['dateTime']));

        $files = [];
        if (isset($action['files'])) {
            foreach ($action['files'] as $file) {
                if (!isset($file[0])) {
                    $file = [$file];
                }

                foreach ($file as $cFile) {
                    $fileData = Yii::$app->planfix->api('file.get', [
                        'file' => [
                            'id' => $cFile['id'],
                        ],
                        'returnDownloadLinks' => true
                    ]);

                    $files[] = $fileData['file']['downloadLink'];
                }
            }
        }
        $aFiles = implode(',', $files);

        if ($aFiles != $dbAction->files) {
            $hasChanges = true;
        }
        $dbAction->files = $aFiles;

        $notify = [];

        if (isset($action['notifiedList']['user']) && isset($action['notifiedList']['user']['id'])) {
            $action['notifiedList']['user'] = [ $action['notifiedList']['user'] ];
        }

        if (isset($action['notifiedList']['user'])) {
            foreach ($action['notifiedList']['user'] as $notifyUser) {
                $notify[] = static::getUserFromPlanfix($notifyUser['id']);

                $user = User::find()->where(['planfix_id' => $notifyUser])->one();
                if ($user && !$user->init_issue_id) {
                    $user->init_issue_id = $dbAction->issue_id;
                    $user->save();
                }
            }
        }

        $notifyList = implode(',', $notify);
        if ($notifyList != $dbAction->notify_list) {
            $hasChanges = true;
        }
        $dbAction->notify_list = $notifyList;

        $save = $dbAction->save();

        $user = $dbAction->creator;
        if (!$user->init_issue_id) {
            $user->init_issue_id = $dbAction->issue_id;
            $user->save();
        }

        if (!$save) {
            throw new Exception(print_r($dbAction->getErrors(), true));
        }

        static::getAnaliticFromAction($action, $issue);

        if ($hasChanges) {
            Yii::$app->gitlab->syncIssueComment($issue, $dbAction);
        }
    }

    /**
     * @param $issue
     * @throws Exception
     */
    public static function syncIssueGitlab($issue)
    {
        $label = explode(',', $issue->labels)[0];
        $status = @static::getPlanfixStatusList(true)[$label]['id'];

        if (!isset(static::getPlanfixStatusList(true)[$label]['id'])) {
            $status = PFApi::getPlanfixStatus($issue->status, $issue->project, true, true);
        }

        $statusMapping = CoreStatus::find()->where([
            'label' => $label,
            'project_id' => $issue->project->id
        ])->one();

        if ($statusMapping) {
            $status = $statusMapping->planfix_id;
        }

        $issueText = Yii::$app->gitlab->renderMarkDown($issue->description_gitlab, $issue->project->gitlab_id);

        if (!$status) {
            $status = Yii::$app->planfix->default_status_id;
        }

        if (!$issue->planfix_id) {
            $task = Yii::$app->planfix->api('task.add', [
                'task' => [
                    'template' => Yii::$app->planfix->tpl_dev_id, // Development template
                    'title' => $issue->title,
                    'description' => $issueText,
                    'status' => $status,
                    'owner' => [
                        ['id' => (string)$issue->creator->planfix_id]
                    ],
                    'project' => [
                        ['id' => (string)$issue->project->planfix_id]
                    ],
                    'endDateIsSet' => $issue->due_date? true: false,
                    'endDate' => $issue->due_date ? date('d-m-Y', strtotime($issue->due_date)) : '',
                    'endTimeIsSet' => $issue->due_date? true: false,
                    'endTime' => $issue->due_date ? date('H:i', strtotime($issue->due_date)) : '',
                    'workers' => [
                        'users' => [
                            ['id' => (string)$issue->assignee->planfix_id]
                        ],
                    ],
                ]
            ], false, true);

            $issue->planfix_id = $task['task']['general'];
            $issue->description = $issueText;
            $issue->status = (string)$status;
            $save = $issue->save();
            if (!$save) {
                throw new Exception("Cant save issue " . print_r($issue->getErrors(), true));
            }
        } else {
            Yii::$app->planfix->api('task.update', [
                'task' => [
                    'general' => $issue->planfix_id,
                    'title' => $issue->title,
                    'description' => $issueText,
                    'status' => $status,
                    'owner' => [
                        ['id' => (string)$issue->creator->planfix_id]
                    ],
                    'project' => [
                        ['id' => (string)$issue->project->planfix_id]
                    ],
                    'endDateIsSet' => $issue->due_date? true: false,
                    'endDate' => $issue->due_date ? date('d-m-Y', strtotime($issue->due_date)) : '',
                    'endTimeIsSet' => $issue->due_date? true: false,
                    'endTime' => $issue->due_date ? date('H:i', strtotime($issue->due_date)) : '',
                    'workers' => [
                        'users' => [
                            ['id' => (string)$issue->assignee->planfix_id]
                        ],
                    ],
                ]
            ], false, true);

            $issue->description = $issueText;
            $issue->status = $status;
            $issue->save();
        }
    }

    /**
     * @param $note
     * @return bool
     * @throws Exception
     */
    public static function syncNoteGitlab($note)
    {
        $analitics = static::getNoteAnalitic($note);
        if (count($analitics)) {
            if (Time::find()->where(['gitlab_id' => $note->gitlab_id])->count()) {
                $analitics = [];
            }
        }
        if ($note->is_log && !count($analitics)) {
            return true;
        }

        $notifiedList = [];

        foreach (explode(',', $note->notify_list) as $id) {
            $user = User::findOne($id);
            if (!$user || !$user->planfix_id) {
                continue;
            }
            $notifiedList[] = ['id' => $user->planfix_id];
        }

        if (!$note->creator->planfix_id) {
            Yii::$app->impulse->air('app_error', 'Пользователь ' . $note->creator->username . ' не найден в Planfix. Комменатрий не будет добавлен');
            return false;
        }

        $noteText = $note->getPlanfixText();

        if (!$note->planfix_id) {
            try {
                $action = Yii::$app->planfix->api('action.add', [
                    'action' => [
                        'description' => $note->is_log? '' : $noteText,
                        'task' => [
                            ['general' => $note->issue->planfix_id]
                        ],
                        'notifiedList' => [
                            'user' => $notifiedList
                        ],
                        'owner' => [
                            'id' => $note->creator->planfix_id,
                        ],
                        'dateTime' => date('d-m-Y H:i', strtotime($note->created_at)),
                        'analitics' => $analitics,
                    ]
                ], false, true);
            } catch (Exception $e) {
                if ($e->getMessage() === 'Указанная задача не существует -> task') {
                    PFApi::syncIssueGitlab($note->issue);

                    $note->issue->refresh();
                    
                    Yii::$app->gitlab->importNotes($note->issue, $note->issue->project);
                    return PFApi::syncNoteGitlab($note);
                }
                throw $e;
            }

            $note->planfix_id = $action['action']['id'];
            $note->text = $noteText;
            $note->save();

            if (count($analitics)) {
                $action = Yii::$app->planfix->api('action.get', [
                    'action' => [
                        'id' => $note->planfix_id
                    ]
                ], false, true);

                static::getAnaliticFromAction($action['action'], $note->issue, $note->gitlab_id);
            }
        } else {
            Yii::$app->planfix->api('action.update', [
                'action' => [
                    'id' => $note->planfix_id,
                    'description' => $note->is_log? '' : $noteText,
                    'notifiedList' => [
                        'user' => $notifiedList
                    ],
                    'owner' => [
                        'id' => $note->creator->planfix_id,
                    ],
                    'dateTime' => date('d-m-Y H:i', strtotime($note->created_at)),
                ]
            ], false, true);

            $note->text = $note->is_log? '' : $noteText;
            $note->save();
        }

        static::checkSetTaskMember($note->creator, $note->issue->planfix_id);

        return true;
    }

    /**
     * @param $user
     * @param $taskId
     * @return bool
     */
    public static function checkSetTaskMember($user, $taskId)
    {
        $task = Yii::$app->planfix->api('task.get', [
            'task' => [
                'general' => $taskId
            ]
        ]);

        if (!isset($task['task']['members']['users'])) {
            $members = [];
        } else {
            $members = $task['task']['members']['users']['user'];
        }

        if (isset($members['id'])) {
            $members = [ $members ];
        }
        
        $tMembers = [];
        foreach ($members as $member) {
            if ($member['id'] === $user->planfix_id) {
                return true;
            }
            $tMembers[] = [ 'id' => (string)$member['id'] ];
        }

        $tMembers[] = [ 'id' => (string)$user->planfix_id ];

        $task = Yii::$app->planfix->api('task.update', [
            'task' => [
                'general' => $taskId,
                'members' => [
                    'users' => $tMembers,
                ],
            ]
        ], false, true);


        return $task;
    }

    /**
     * @param $note
     * @return array
     */
    public static function getNoteAnalitic($note)
    {
        $time = Yii::$app->gitlab->getCommentTime($note);

        if ($time) {
            $date = date('d-m-Y 00:00:00', strtotime($time['date']));

            if ($date == date('d-m-Y 00:00:00', strtotime($note->created_at))) {
                $date = date('d-m-Y H:i:s', strtotime($note->created_at));
            }

            $day = date('d-m-Y', strtotime($date));
            $amount = $time['amount'];

            $analitics = [];

            while ($amount > 0) {
                $dayStart = strtotime("$day 00:00:00");

                if (strtotime($date) - $amount >= $dayStart) {
                    $data = [
                        [
                            'itemData' => [
                                'fieldId' => Yii::$app->planfix->time_analytic_date_id, // Date
                                'value' => date('d-m-Y', $dayStart)
                            ]
                        ],
                        [
                            'itemData' => [
                                'fieldId' => Yii::$app->planfix->time_analytic_time_id,
                                'value' => [
                                    'begin' => date('H:i', strtotime($date) - $amount),
                                    'end' => date('H:i', strtotime($date)),
                                ]
                            ]
                        ],
                        
                    ];

                    if (Yii::$app->planfix->time_analytic_user_id) {
                        $data[] = [
                            'itemData' => [
                                'fieldId' => Yii::$app->planfix->time_analytic_user_id, // User
                                'value' => [
                                    'id' => $note->creator->planfix_id
                                ]
                            ]
                        ];
                    }

                    $analitics[] = [
                        'analitic' => [
                            'id' => Yii::$app->planfix->time_analytic_id,
                            'analiticData' => $data
                        ]
                    ];

                    $amount = 0;
                } else {
                    $cAmount = strtotime($date) - $dayStart;
                    
                    if ($cAmount > $amount) {
                        $cAmount = $amount;
                    }
                    $amount -=$cAmount;

                    $data = [
                        [
                            'itemData' => [
                                'fieldId' => Yii::$app->planfix->time_analytic_date_id, // Date
                                'value' => date('d-m-Y', $dayStart)
                            ]
                        ],
                        [
                            'itemData' => [
                                'fieldId' =>  Yii::$app->planfix->time_analytic_time_id,
                                'value' => [
                                    'begin' => date('H:i', strtotime($date) - $cAmount),
                                    'end' => date('H:i', strtotime($date)),
                                ]
                            ]
                        ]
                    ];


                    if (Yii::$app->planfix->time_analytic_user_id) {
                        $data[] = [
                            'itemData' => [
                                'fieldId' => Yii::$app->planfix->time_analytic_user_id, // User
                                'value' => [
                                    'id' => $note->creator->planfix_id
                                ]
                            ]
                        ];
                    }

                    $analitics[] = [
                        'analitic' => [
                            'id' => Yii::$app->planfix->time_analytic_id,
                            'analiticData' => $data,
                        ]
                    ];

                    $day = date('d-m-Y', strtotime('-1 day', strtotime($day)));
                    $date = date('d-m-Y H:i:s', strtotime($date) - $cAmount);
                }
            }

            return $analitics;
        }

        return [];
    }
}
