<?php

namespace app\planfix\models;

use app\core\models\Project;
use app\core\models\User;
use Yii;
use yii\behaviors\TimestampBehavior;
use yii\db\ActiveRecord;
use yii\db\Expression;

/**
 * This is the model class for table "planfix_work_analytic".
 *
 * @property int    $id
 * @property int    $pf_action_id
 * @property int    $pf_task_id
 * @property int    $pf_author_id
 * @property int    $pf_analytic_id
 * @property int    $user_id
 * @property int    $start_date
 * @property int    $end_date
 * @property int    $amount
 * @property string $created_at
 * @property string $updated_at
 * @property int    $type_id
 * @property string $result
 */
class PlanfixWorkAnalytic extends ActiveRecord
{
	/**
	 * @inheritdoc
	 */
	public static function tableName()
	{
		return 'planfix_work_analytic';
	}

	/**
	 * @inheritdoc
	 */
	public function rules()
	{
		return [
			[
				[
					'pf_action_id',
					'pf_task_id',
					'pf_author_id',
					'pf_analytic_id',
					'project_id',
					'user_id',
					'start_date',
					'end_date',
					'type_id',
				],
				'required',
			],
			[['pf_action_id', 'pf_task_id', 'pf_author_id', 'user_id', 'type_id'], 'integer'],
			[['start_date', 'end_date', 'created_at', 'updated_at', 'result', 'amount'], 'safe'],
		];
	}

	/**
	 * @inheritdoc
	 */
	public function attributeLabels()
	{
		return [
			'id'             => 'ID',
			'pf_action_id'   => 'ID действия',
			'pf_task_id'     => 'ID задачи',
			'pf_author_id'   => 'ID автора',
			'pf_analytic_id' => 'Ключ аналитики',
			'project_id'     => 'ID проекта',
			'user_id'        => 'ID пользователя',
			'start_date'     => 'Время начала работы',
			'end_date'       => 'Время окончания работы',
			'amount'         => 'Потраченное время',
			'created_at'     => 'Дата создания',
			'updated_at'     => 'Дата обновления',
			'type_id'        => 'Вид работ',
			'result'         => 'Результат',
		];
	}

	/**
	 * @inheritdoc
	 */
	public function behaviors()
	{
		return [
			[
				'class' => TimestampBehavior::class,
				'value' => new Expression('NOW()'),
			],
		];
	}

	/**
	 * @inheritDoc
	 */
	public function beforeSave($insert)
	{
		// Посчитать потраченное время
		$this->onRecalcAmountTime();

		return parent::beforeSave($insert);
	}

	public static $typesOfWork = [
		"20" => "Исправление / Доработка кода",
		"60" => "Диалог с поддержкой",
		"53" => "Настройка Планфикса",
		"58" => "Переговоры с Заказчиком",
		"62" => "Проверка текущих настроек",
		"55" => "Создание схемы процесса BPMN",
		"57" => "Составление плана работ",
		"8"  => "Админ. работа",
		"63" => "Диалог с коллегами",
		"50" => "Менеджерское сопровождение",
		"51" => "Обучение сотрудника Заказчика",
		"9"  => "nbsp;Обучение сотрудников",
		"64" => "Оформление и проверка документов",
		"48" => "Дизайн",
		"47" => "Доработка изображений",
		"4"  => "Доработка дизайн-макета",
		"3"  => "Подготовка макета к печати",
		"40" => "Поиск идеи",
		"56" => "Покупка темы",
		"2"  => "Разработка концепции",
		"5"  => "Разработка дизайна",
		"6"  => "Флеш-анимация",
	];

	/**
	 * Вернуть вид работ
	 *
	 * @param $type_id
	 *
	 * @return int|mixed
	 */
	public static function getTypeOfWork($type_id)
	{
		if (self::checkTypeOfWork($type_id)) {
			return self::$typesOfWork[$type_id];
		}

		return 0;
	}

	/**
	 * Проверить вид работ
	 *
	 * @param $type_id
	 *
	 * @return int|mixed
	 */
	public static function checkTypeOfWork($type_id)
	{
		return array_key_exists($type_id, self::$typesOfWork);
	}

	/**
	 * Посчитать потраченное время
	 *
	 * @param bool $save
	 */
	public function onRecalcAmountTime($save = false)
	{
		$this->amount = strtotime($this->end_date) - strtotime($this->start_date);
	}

	/**
	 * Связь аналитики с пользователем в системе
	 *
	 * @return \yii\db\ActiveQuery
	 */
	public function getUser()
	{
		return $this->hasOne(User::class, ['id' => 'user_id']);
	}

	/**
	 * Связь аналитики с проектом в системе
	 *
	 * @return \yii\db\ActiveQuery
	 */
	public function getProject()
	{
		return $this->hasOne(Project::class, ['id' => 'project_id']);
	}

	/**
	 * Вернуть отформатированное время
	 *
	 * @return string
	 */
	public function timeToRaw()
	{
		return Yii::$app->formatter->asDuration($this->amount);
	}

	/**
	 * Ссылка на задачу в планфикса
	 *
	 * @return string
	 */
	public function getPlanfixTaskUrl()
	{
		return "https://poravinternet.planfix.ru/?action=task&id=" . $this->pf_task_id;
	}

	/**
	 * Ссылка на действие в задаче планфикса
	 *
	 * @return string
	 */
	public function getPlanfixActionUrl()
	{
		return $this->getPlanfixTaskUrl() . "&comment=" . $this->pf_action_id;
	}
}
